@@ -2,9 +2,9 @@
 # handling GnuPG ed25519 ECC keys.
 #
 # Licensed under the GPLv3: http://www.gnu.org/licenses/gpl-3.0.html
-# Copyright © 2019-2020 Gonzo Investigatory Journalism Agency, LLC
+# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
 #             <gonzo.development@protonmail.ch>
-#           © 2019-2020 Richard Machado <rmlibre@riseup.net>
+#           © 2019-2021 Richard Machado <rmlibre@riseup.net>
 # All rights reserved.
 #
 
@@ -25,17 +25,18 @@ tiny_gnupg instead of the --user flag.
 import sys
 import pytest
 import asyncio
-import subprocess
 from pathlib import Path
 from aiohttp import ClientSession
-from aiohttp_socks import SocksConnector
+from aiohttp_socks import ProxyConnector
+
 
 PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
 sys.path.append(PACKAGE_PATH)
-new_task = asyncio.get_event_loop().create_task
+
 
 from tiny_gnupg import GnuPG, run
 
+
 legacy_key = """-----BEGIN PGP PUBLIC KEY BLOCK-----
 Version: OpenPGP.js v4.3.0
 Comment: https://openpgpjs.org
@@ -142,16 +143,28 @@ PzNx/ZogrbmOMOWsTCC1DFycAjCRtueMDG83FVl0zdX0fFR5cZnRdxde+Y6oyZ6u
 """
 
 
+_homedir = Path(PACKAGE_PATH).absolute() / "tiny_gnupg/gpghome"
+_options = _homedir / "gpg2.conf"
+_executable = "/usr/bin/gpg2"
+
+
 @pytest.fixture(scope="module")
 def gpg():
     print("setup".center(15, "-"))
+
     username = "testing_user"
     email = "testing_user@testing.testing"
     passphrase = "test_passphrase"
-    relative_gpg_path = str(Path(PACKAGE_PATH).absolute() / "tiny_gnupg/gpghome")
-    gpg = GnuPG(username, email, passphrase)
-    gpg.set_homedir(relative_gpg_path)
-    gpg.reset_daemon()
+
+    gpg = GnuPG(
+        email=email,
+        username=username,
+        passphrase=passphrase,
+        homedir=_homedir,
+        options=_options,
+        executable=_executable,
+    )
+
     yield gpg
     print("teardown".center(18, "-"))
 
@@ -161,8 +174,8 @@ def pop(dictionary):
 
 
 async def async_method_runner(gpg):
-    assert gpg.Connector.__class__ == SocksConnector
-    async with gpg.Session as session:
+    assert gpg.network._Connector.__class__ == ProxyConnector
+    async with gpg.network._Session as session:
         session.__class__ == ClientSession
 
 
@@ -172,45 +185,56 @@ def test_instance(gpg):
             gpg.delete(gpg.email)
         except:
             break
-    gpg.set_home_permissions("/ridiculous_root_directory_not_real")
-    # Successfully failed to change permissions on an invalid dir.
-    # We refrain from testing on a root folder that may actually exist
-    # for safety of the tester reasons.
+    try:
+        gpg._set_homedir_permissions("/ridiculous_root_directory_not_real")
+    except FileNotFoundError:
+        "Successfully failed to change permissions on an invalid dir. "
+        "We refrain from testing on, or making, a priveleged folder "
+        "that may actually exist for safety of the tester reasons."
+    else:
+        raise AssertionError(
+            "The instance was able to change the fake directory's "
+            "permissions."
+        )
+
     gpg.gen_key()
-    test_gpg = GnuPG(gpg.username, gpg.email, gpg.passphrase)
+    test_gpg = GnuPG(
+        email=gpg.user.email,
+        username=gpg.user.username,
+        passphrase=gpg.user.passphrase,
+        homedir=_homedir,
+        options=_options,
+        executable=_executable,
+    )
     test_gpg.gen_key()
-    assert gpg.username == test_gpg.username
-    assert gpg.email == test_gpg.email
-    assert gpg.passphrase == test_gpg.passphrase
-    assert gpg.port == test_gpg.port
-    assert gpg.tor_port == test_gpg.tor_port
-    assert gpg.home == test_gpg.home
+    assert gpg.user.username == test_gpg.user.username
+    assert gpg.user.email == test_gpg.user.email
+    assert gpg.user.passphrase == test_gpg.user.passphrase
+    assert gpg.network.port == test_gpg.network.port
+    assert gpg.network.tor_port == test_gpg.network.tor_port
+    assert gpg.homedir == test_gpg.homedir
     assert gpg.executable == test_gpg.executable
-    assert gpg._Connector == test_gpg._Connector
-    assert gpg._Session == test_gpg._Session
     run(async_method_runner(gpg))
-    assert gpg._search_string == test_gpg._search_string
-    assert gpg.keyserver == test_gpg.keyserver
-    assert str(gpg.port) in gpg.keyserver
-    assert gpg.keyserver_export_api == test_gpg.keyserver_export_api
-    assert gpg.keyserver_verify_api == test_gpg.keyserver_verify_api
-    assert gpg.searchserver == test_gpg.searchserver
-    assert gpg.base_command == test_gpg.base_command
-    assert gpg.base_passphrase_command == test_gpg.base_passphrase_command
+    assert gpg._search_prefix == test_gpg._search_prefix
+    assert gpg._keyserver == test_gpg._keyserver
+    assert str(gpg.network.port) in gpg._keyserver
+    assert gpg._keyserver_export_api == test_gpg._keyserver_export_api
+    assert gpg._keyserver_verify_api == test_gpg._keyserver_verify_api
+    assert gpg._searchserver == test_gpg._searchserver
+    assert gpg._base_command == test_gpg._base_command
+    assert gpg._base_passphrase_command == test_gpg._base_passphrase_command
     ###
-    assert gpg.username == "testing_user"
-    assert gpg.email == "testing_user@testing.testing"
+    assert gpg.network.port == 80
+    assert gpg.network.tor_port == 9050
+    assert ".onion" in gpg._keyserver
+    assert ".onion" in gpg._searchserver
     assert len(gpg.fingerprint) == 40
     assert type(gpg.fingerprint) == str
-    assert gpg.passphrase == "test_passphrase"
-    assert gpg.home.endswith("gpghome")
+    assert gpg.homedir.endswith("gpghome")
+    assert gpg.user.username == "testing_user"
     assert gpg.executable.endswith("gpg2")
-    assert gpg._Connector == SocksConnector
-    assert gpg._Session == ClientSession
-    assert gpg.port == 80
-    assert gpg.tor_port == 9050
-    assert ".onion" in gpg.keyserver
-    assert ".onion" in gpg.searchserver
+    assert gpg.user.passphrase == "test_passphrase"
+    assert gpg.user.email == "testing_user@testing.testing"
     assert gpg.fingerprint in gpg.list_keys()
     assert gpg.fingerprint in gpg.list_keys(secret=True)
     assert test_gpg.fingerprint in gpg.list_keys()
@@ -232,24 +256,98 @@ def test_encode_inputs(gpg):
 
 def test_command(gpg):
     options = ["--list-keys"]
-    command = gpg.command(*options)
-    passphrase_command = gpg.command(*options, with_passphrase=True)
+    command = gpg.encode_command(*options)
+    passphrase_command = gpg.encode_command(*options, with_passphrase=True)
     for option in options:
         assert option in command
         assert option in passphrase_command
         command.remove(option)
-        assert command == gpg.command()
+        assert command == gpg.encode_command()
         passphrase_command.remove(option)
-        assert passphrase_command == gpg.command(with_passphrase=True)
+        assert passphrase_command == gpg.encode_command(with_passphrase=True)
+
+
+def test_manual_command(gpg):
+    custom_command = gpg.encode_command(manual=True)
+    default_command = gpg.encode_command(manual=False)
+    default_command.remove("--yes")
+    default_command.remove("--batch")
+    default_command.remove("--no-tty")
+    assert custom_command == default_command
+    default_passphrase_command = gpg.encode_command(with_passphrase=True)
+    custom_passphrase_command = gpg.encode_command(
+        with_passphrase=True, manual=True
+    )
+    assert default_passphrase_command == custom_passphrase_command
 
 
 def test_export_import(gpg):
     secret_key = gpg.text_export(gpg.fingerprint, secret=True)
     gpg.text_import(secret_key)
     try:
+        failed = False
         gpg.text_export(gpg.fingerprint, secret="Non-boolean value")
     except:
+        failed = True
         """Successfully blocked non-boolean"""
+    finally:
+        assert failed
+
+
+def test_isolated_identities(gpg):
+    anon = GnuPG(
+        username="anon_user",
+        email="anonymous@testing.testing",
+        passphrase="rubbish_pw",
+        executable=_executable,
+    )
+    anon.gen_key()
+    anon_uid = anon.fingerprint
+    ### decrypting
+    enc_msg = gpg.encrypt("hi!", anon_uid)
+    msg = anon.decrypt(enc_msg)
+    try:
+        failed = False
+        gpg.decrypt(enc_msg)
+    except LookupError:
+        failed = True
+        """
+        Same user succefully prevented from decrypting message sent to a
+        different identity.
+        """
+    finally:
+        assert failed
+    ### signatures
+    sig_0 = anon.sign("message", local_user=anon_uid)
+    sig_1 = anon.sign("message")
+    try:
+        failed = False
+        gpg.sign("message", local_user=anon_uid)
+    except PermissionError:
+        failed = True
+        """
+        Same user successfully prevented from signing message with the
+        secret key associated with another identity.
+        """
+    finally:
+        assert failed
+    ### encrypting
+    try:
+        failed = False
+        gpg.encrypt("greetings", anon_uid, local_user=anon_uid)
+    except PermissionError:
+        failed = True
+        """
+        Same user successfully prevented from signing an encrypted message
+        with the secret key associated with another identity.
+        """
+    finally:
+        assert failed
+    while True:
+        try:
+            anon.delete(anon.email)
+        except:
+            break
 
 
 def test_cipher(gpg):
@@ -259,7 +357,7 @@ def test_cipher(gpg):
     for trust_level in range(0, 7):
         for fingerprint in gpg.list_keys():
             try:
-                gpg.trust(fingerprint, trust_level)
+                gpg.set_key_trust(fingerprint, trust_level)
             except ValueError as invalid_trust_level:
                 # if 1 > int(trust_level) or 5 < int(trust_level):
                 """Successfully blocked invlaid trust level"""
@@ -289,8 +387,12 @@ def test_cipher(gpg):
         signed_message_1 = gpg.sign(signed_message_0)
         try:
             gpg.sign(message, key="Non boolean value")
+            blocked_non_boolean = False
         except:
             """Successfully blocked non-boolean value"""
+            blocked_non_boolean = True
+        else:
+            assert blocked_non_boolean
         gpg.verify(signed_message_0)
         gpg.verify(signed_message_1)
     ###
@@ -299,7 +401,7 @@ def test_cipher(gpg):
         msg = encrypted_message_0
         corrupt_message = msg[:201] + msg[202:]
         gpg.decrypt(corrupt_message)
-    except subprocess.CalledProcessError:
+    except TypeError:
         failed = True
         # The metadata on the message is corrupted by dropping a byte,
         # expectedly leading to an error.
@@ -309,14 +411,20 @@ def test_cipher(gpg):
     username = "test_sender"
     email = "test_sender@testing.testing"
     passphrase = "test_sender_passphrase"
-    sender = GnuPG(username, email, passphrase)
+
+    sender = GnuPG(
+        email=email,
+        username=username,
+        passphrase=passphrase,
+        executable=_executable,
+    )
     sender.gen_key()
     sender_key = sender.list_keys(sender.fingerprint)
     sender_pkey = sender.text_export(sender.fingerprint)
     sender_skey = sender.text_export(sender.fingerprint, secret=True)
     msg = sender.encrypt("testing", gpg.fingerprint)
     sender.delete(sender.fingerprint)
-    sender.reset_daemon()
+    sender._reset_daemon()
     try:
         failed = False
         gpg.decrypt(msg)
@@ -337,11 +445,11 @@ def test_cipher(gpg):
 
 
 def test_file_io(gpg):
-    path = Path(gpg.home).absolute()
+    path = Path(gpg.homedir).absolute()
     file_path = str(path / f"{gpg.fingerprint}.asc")
     key = gpg.text_export(gpg.fingerprint)
-    run(gpg.file_export(path, gpg.fingerprint))
-    run(gpg.file_import(file_path))
+    gpg.file_export(path, gpg.fingerprint)
+    gpg.file_import(file_path)
     Path(file_path).unlink()
 
 
@@ -353,12 +461,12 @@ def test_networking(gpg):
     assert " " not in key_url
     assert "<" not in key_url
     assert ">" not in key_url
-    key = run(gpg.get(key_url))
+    key = run(gpg.network.get(key_url))
     gpg.text_import(key)
     assert gpg.list_keys(dev_email)
     fingerprint = gpg.key_fingerprint(dev_email)
     assert dev_fingerprint == fingerprint
-    assert fingerprint == pop(gpg.list_keys(dev_fingerprint))
+    assert fingerprint in gpg.list_keys(dev_fingerprint)
     email = gpg.key_email(fingerprint)
     assert dev_email == email
     assert email == gpg.list_keys(dev_email)[fingerprint]
@@ -366,7 +474,7 @@ def test_networking(gpg):
     key_from_fingerprint = gpg.text_export(fingerprint)
     assert key_from_email == key_from_fingerprint
     gpg.text_import(key_from_email)
-    run(gpg.network_import(dev_email))
+    # run(gpg.network_import(dev_email))
     try:
         assert key == key_from_email
     except:
@@ -376,7 +484,7 @@ def test_networking(gpg):
     run(gpg.network_export(gpg.fingerprint))
     test_key_url = run(gpg.search(gpg.fingerprint))
     local_key = gpg.text_export(gpg.fingerprint)
-    network_key = run(gpg.get(test_key_url))
+    network_key = run(gpg.network.get(test_key_url))
     assert local_key != network_key  # removed and/or reencoded uids
     try:
         failed = False
@@ -404,7 +512,6 @@ def test_networking(gpg):
         """Successfully failed to retrieve data for bogus query"""
 
 
-
 def test_network_concurrency(gpg):
     async def gather_looper(gpg, uid):
         tasks = await looper(gpg, uid)
@@ -413,7 +520,7 @@ def test_network_concurrency(gpg):
     async def looper(gpg, uid):
         tasks = []
         for i in range(2):
-            tasks.append(new_task(gpg.search(uid)))
+            tasks.append(asyncio.ensure_future(gpg.search(uid)))
         return tasks
 
     uid = "support@keys.openpgp.org"
@@ -422,20 +529,17 @@ def test_network_concurrency(gpg):
     assert url.strip()
     for link in urls:
         assert url == link
-    urls = run(gather_looper(gpg, uid))
-    for link in urls:
-        assert url == link
 
 
 def test_key_signing(gpg):
     dev_email = "gonzo.development@protonmail.ch"
     dev_fingerprint = "31FDCC4F9961AFAC522A9D41AE2B47FA1EF44F0A"
-    command = gpg.command("--check-sigs")
+    command = gpg.encode_command("--check-sigs")
     keyring = gpg.read_output(command)
     gpg.sign(dev_fingerprint, key=True)
-    signed_keying = gpg.read_output(command)
-    assert keyring != signed_keying
-    condensed_keyring = signed_keying.replace(" ", "")
+    signed_keyring = gpg.read_output(command)
+    assert keyring != signed_keyring
+    condensed_keyring = signed_keyring.replace(" ", "")
     fingerprint = gpg.fingerprint[-16:]
     assert f"<{dev_email}>\nsig!0x{fingerprint}" in condensed_keyring
 
@@ -446,14 +550,14 @@ def test_packet_parsing(gpg):
     encrypted_message = gpg.encrypt("test", gpg.fingerprint, sign=False)
     gpg_key = gpg.text_export(gpg.fingerprint)
     ###
-    signature_fingerprint = gpg.packet_fingerprint(signature)
-    signed_encrypted_message_fingerprint = gpg.packet_fingerprint(
+    signature_fingerprint = gpg._packet_fingerprint(signature)
+    signed_encrypted_message_fingerprint = gpg._packet_fingerprint(
         signed_encrypted_message
     )
-    encrypted_message_fingerprint = gpg.packet_fingerprint(
+    encrypted_message_fingerprint = gpg._packet_fingerprint(
         encrypted_message
     )
-    gpg_key_fingerprint = gpg.packet_fingerprint(gpg_key)
+    gpg_key_fingerprint = gpg._packet_fingerprint(gpg_key)
     ###
     key = gpg.list_keys(gpg.fingerprint)
     key_from_signature = gpg.list_keys(signature_fingerprint)
@@ -488,13 +592,13 @@ def test_auto_fetch_methods(gpg):
     msg = run(gpg.auto_decrypt(dev_signed_message))
     dev_key = gpg.text_export(dev_fingerprint)
     gpg.text_import(legacy_key)
-    assert msg.strip() == message
+    assert msg == message
     ###
-    packets_0 = gpg.list_packets(dev_signed_encrypted_message)
-    packets_1 = gpg.list_packets(dev_encrypted_message)
-    packets_2 = gpg.list_packets(dev_signed_message)
-    packets_3 = gpg.list_packets(legacy_key)
-    packets_4 = gpg.list_packets(dev_key)
+    packets_0 = gpg._list_packets(dev_signed_encrypted_message)
+    packets_1 = gpg._list_packets(dev_encrypted_message)
+    packets_2 = gpg._list_packets(dev_signed_message)
+    packets_3 = gpg._list_packets(legacy_key)
+    packets_4 = gpg._list_packets(dev_key)
     assert type(packets_0) == list
     assert type(packets_1) == list
     assert type(packets_2) == list
@@ -506,15 +610,15 @@ def test_auto_fetch_methods(gpg):
     assert len(packets_3) == 28
     assert len(packets_4) == 31
     try:
-        gpg.list_packets(20 * "Non-OpenPGP data")
+        gpg._list_packets(20 * "Non-OpenPGP data")
     except:
         """Successfully failed when invalid data sent for parsing"""
     ###
-    fingerprint_0 = gpg.packet_fingerprint(dev_signed_encrypted_message)
-    fingerprint_1 = gpg.packet_fingerprint(dev_encrypted_message)
-    fingerprint_2 = gpg.packet_fingerprint(dev_signed_message)
-    fingerprint_3 = gpg.packet_fingerprint(legacy_key)
-    fingerprint_4 = gpg.packet_fingerprint(dev_key)
+    fingerprint_0 = gpg._packet_fingerprint(dev_signed_encrypted_message)
+    fingerprint_1 = gpg._packet_fingerprint(dev_encrypted_message)
+    fingerprint_2 = gpg._packet_fingerprint(dev_signed_message)
+    fingerprint_3 = gpg._packet_fingerprint(legacy_key)
+    fingerprint_4 = gpg._packet_fingerprint(dev_key)
     fingerprint_0_key = gpg.list_keys(fingerprint_0)
     fingerprint_1_key = gpg.list_keys(fingerprint_1)
     fingerprint_2_key = gpg.list_keys(fingerprint_2)
@@ -544,8 +648,8 @@ def test_auto_fetch_methods(gpg):
         # on an encrypted message in general, unless the message is
         # specifcally a signature, not encrypted plaintext. This is just
         # not how verify works. Signatures are on the inside on encrypted
-        # messages. So ``decrypt()`` should be used, it throws if a
-        # signature is invalid on a message.
+        # messages. So ``decrypt()`` should be used instead, it throws if
+        # a signature is invalid on a message.
     try:
         failed = False
         run(gpg.auto_verify(dev_encrypted_message))
@@ -561,7 +665,7 @@ def test_auto_fetch_methods(gpg):
 
 
 def test_revoke(gpg):
-    raw_list_keys = gpg.raw_list_keys(gpg.fingerprint).replace(" ", "")
+    raw_list_keys = gpg._raw_list_keys(gpg.fingerprint).replace(" ", "")
     assert "[revoked]" not in raw_list_keys
     try:
         failed = False
@@ -572,7 +676,7 @@ def test_revoke(gpg):
         assert failed  # GnuPG bug #T4393
         # this should not fail when the bug is resolved
     gpg.revoke(gpg.fingerprint)
-    raw_list_keys = gpg.raw_list_keys(gpg.fingerprint).replace(" ", "")
+    raw_list_keys = gpg._raw_list_keys(gpg.fingerprint).replace(" ", "")
     assert "[revoked]" in raw_list_keys
     run(gpg.network_export(gpg.fingerprint))
     try:
@@ -581,23 +685,15 @@ def test_revoke(gpg):
     except:
         failed = True
     finally:
-        assert failed  # server removes the key after revocation? No.
-        # See https://gitlab.com/hagrid-keyserver/hagrid/issues/137
-        # GnuPG bug #T4393 will cause crash
+        assert failed  # server removes the key's uid information after
+        # revocation now. But this test also fails because of a known
+        # bug in GnuPG: bug #T4393 will cause crash.
 
 
 def test_delete(gpg):
-    dev_email = "gonzo.development@protonmail.ch"
-    email = "testing_user@testing.testing"
-    amount_of_test_keys = 0
-    for key_email in gpg.list_keys().values():
-        if key_email == email:
-            amount_of_test_keys += 1
+    number_of_keys_before_delete = len(gpg.list_keys())
     gpg.delete(gpg.fingerprint)
-    amount_of_test_keys_after_delete = 0
-    for key_email in gpg.list_keys().values():
-        amount_of_test_keys_after_delete += 1 if key_email == email else 0
-    assert amount_of_test_keys - 1 == amount_of_test_keys_after_delete
+    assert len(gpg.list_keys()) + 1 == number_of_keys_before_delete
     while True:
         try:
             gpg.delete(gpg.email)
@@ -621,4 +717,5 @@ def test_delete(gpg):
 
 
 def test_reset_daemon(gpg):
-    gpg.reset_daemon()
+    gpg._reset_daemon()
+
